/*
   decode.c

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: decode.c,v 1.1.1.1 2004/12/08 02:44:16 jleplast Exp $ */

#include <stdio.h>
#include <assert.h>

#include "decode.h"

#define _BYTE_ORDER _LITTLE_ENDIAN

/* Unlike zillions of existing 'typecast' packet decoder
   implementations, this one treats incoming packet data
   as a stream.
   This implementation is probably not the fastest one,
   but it allows to completely avoid inclusion of numerous
   system headers and/or yet another definition of
   ether/ip/tcp/udp/icmp packet structure.
*/

union {	/* this union is to extract 'l' and 'h' portions of the byte value */
  u_char v;
  struct {
#if _BYTE_ORDER   == _LITTLE_ENDIAN
    u_int8_t h:4;               /* header length */
    u_int8_t l:4;                /* version */
#elif _BYTE_ORDER == _BIG_ENDIAN
    u_int8_t l:4;                /* version */
    u_int8_t h:4;               /* header length */
#else
# error "need to define _BYTE_ORDER"
#endif
  } s;
} byte;

typedef struct {     /* packet + integers to track current position */
  char *p;
  int cur;
  int total;
  int saved;
} pbuf_t;

/* Prototypes
 */

/* decoding functions
 */
static int ether_info(packet_info_t *, pbuf_t *);
static int ip_info(packet_info_t *, pbuf_t *);
static int tcp_info(packet_info_t *, pbuf_t *);
static int udp_info(packet_info_t *, pbuf_t *);
static int icmp_info(packet_info_t *, pbuf_t *);

/* stream handling functions
 */
static inline char *_next(pbuf_t *b, int n);
static inline u_int8_t next_8(pbuf_t *b);
static inline u_int16_t next_16(pbuf_t *b);
static inline u_int32_t next_32(pbuf_t *b);

static inline int skip(pbuf_t *b, int n);
static inline int getlen(pbuf_t *b);
static inline char *getbuf(pbuf_t *b);
static inline void savepos(pbuf_t *b);
static inline int getdelta(pbuf_t *b);

/* Implementations
 */
static int ether_info(packet_info_t *info, pbuf_t *p)
{
  if(getlen(p) < 14) return 0;	/* size of ETHERNET packet */

  skip(p, 12);			/* skip HW addresses */

  info->ether.proto = ntohs(next_16(p));
  info->plen = getlen(p);
  info->payload = getbuf(p);

  return 1;
}

static int ip_info(packet_info_t *info, pbuf_t *p)
{
  u_short offset;

  if(getlen(p) < 20) return 0; /* size of IP structure without IP options */
  savepos(p);

  info->is_ip = 1;
  
  byte.v = next_8(p);
  info->ip.ver = byte.s.l;
  info->ip.hl  = byte.s.h;

  if(info->ip.ver != 4) printf("IP version != 4, something's screwed\n");

  info->ip.tos = next_8(p);
  info->ip.len = ntohs(next_16(p));
  info->ip.id  = ntohs(next_16(p));

  offset = ntohs(next_16(p));
  info->ip.off =   offset & 0x1fff;
  info->ip.flags = (offset & 0xE000) >> 13;
  info->ip.ub =    (offset & 0x8000) >> 15;
  info->ip.df =    (offset & 0x4000) >> 14;
  info->ip.mf =    (offset & 0x2000) >> 13;

  info->ip.ttl =   next_8(p);
  info->ip.proto = next_8(p);
  info->ip.csum  = ntohs(next_16(p));

  info->ip.sip = ntohl(next_32(p));
  info->ip.dip = ntohl(next_32(p));

  if(info->ip.hl > 5) {		/* got IP options */
    int siz = (info->ip.hl << 2) - 20; /* size of options array */
    if(getlen(p) >= siz) {
      int EOL = 0;
      for(info->ip.opt_count = 0; siz > 0 & !EOL; info->ip.opt_count++) {
	int code = info->ip.opts[info->ip.opt_count].code = next_8(p);
	if(code == 0) EOL = 1;	      /* IPOPT_EOL */
	else if(code == 1) siz -= 1;  /* IPOPT_NOP */
	else {
	  int optlen = next_8(p);
	  if(optlen <= 0 || optlen > siz) break;

	  siz -= optlen;
	  info->ip.opts[info->ip.opt_count].len = optlen;
	  memcpy(info->ip.opts[info->ip.opt_count].data, getbuf(p), optlen - 2);
	  skip(p, optlen - 2);
	}
      }
    }
  }

  {
    int iphead_len = info->ip.hl << 2;
    int payload_len = info->ip.len - iphead_len;
    p->cur = p->saved + iphead_len;
    info->payload = getbuf(p);
    if(payload_len < info->plen) info->plen = payload_len;
  }
  return 1;
}

static int tcp_info(packet_info_t *info, pbuf_t *p)
{
  if(getlen(p) < 20) return 0;
  savepos(p);

  info->is_tcp = 1;

  info->tcp.sp =       ntohs(next_16(p));
  info->tcp.dp =       ntohs(next_16(p));
  info->tcp.seq =      ntohl(next_32(p));
  info->tcp.ack =      ntohl(next_32(p));

  byte.v = next_8(p);
  info->tcp.off =      byte.s.l;
  info->tcp.x2  =      byte.s.h;

  info->tcp.flags =    next_8(p);
  info->tcp.win =      ntohs(next_16(p));
  info->tcp.csum =     ntohs(next_16(p));
  info->tcp.urg =      ntohs(next_16(p));

  if(info->tcp.off > 5) {		/* got TCP options */
    int siz = (info->tcp.off << 2) - 20; /* size of options array */
    if(getlen(p) >= siz) {
      int EOL = 0;
      for(info->tcp.opt_count = 0; siz > 0 & !EOL; info->tcp.opt_count++) {
	int code = info->tcp.opts[info->tcp.opt_count].code = next_8(p);
	if(code == 0) EOL = 1;	      /* IPOPT_EOL */
	else if(code == 1) siz -= 1;  /* IPOPT_NOP */
	else {
	  int optlen = next_8(p);
	  if(optlen <= 0 || optlen > siz) break;

	  siz -= optlen;
	  info->tcp.opts[info->tcp.opt_count].len = optlen;
	  memcpy(info->tcp.opts[info->tcp.opt_count].data, getbuf(p), optlen - 2);
	  skip(p, optlen - 2);
	}
      }
    }
  }

  {
    int tcphead_len = info->tcp.off << 2;
    p->cur = p->saved + tcphead_len;
    info->payload = getbuf(p);
    info->plen -= tcphead_len;
  }

  return 1;
}

static int udp_info(packet_info_t *info, pbuf_t *p)
{
  if(getlen(p) < 8) return 0;
  savepos(p);

  info->is_udp = 1;

  info->udp.sp =       ntohs(next_16(p));
  info->udp.dp =       ntohs(next_16(p));
  info->udp.len =      ntohs(next_16(p));
  info->udp.csum =     ntohs(next_16(p));
 
  info->payload = getbuf(p);
  info->plen -= getdelta(p);

  return 1;
}

static int icmp_info(packet_info_t *info, pbuf_t *p)
{
  if(getlen(p) < 4) return 0;
  savepos(p);

  info->is_icmp = 1;

  info->icmp.type = next_8(p);
  info->icmp.code = next_8(p);
  info->icmp.csum = ntohs(next_16(p));

  switch(info->icmp.type) {
  case 0:			/* echo reply */
  case 8:			/* echo request */
  case 13:			/* timestamp request */
  case 14:			/* timestamp reply */
  case 17:			/* address mask request */
  case 18:			/* address mask reply */
    info->icmp.ext = 1;
    info->icmp.id =  htons(next_16(p));
    info->icmp.seq = htons(next_16(p));
  }

  info->payload = getbuf(p);
  info->plen -= getdelta(p);

  return 1;
}

static inline char *_next(pbuf_t *b, int n)
{
  char *p = b->p + b->cur;
  b->cur += n;
  assert(b->total >= b->cur);
  return p;
}

static inline u_int8_t next_8(pbuf_t *b)
{
  u_int8_t *a = (u_int8_t *) _next(b, 1);
  return *a;
}

static inline u_int16_t next_16(pbuf_t *b)
{
  u_int16_t *a = (u_int16_t *) _next(b, 2);
  return *a;
}

static inline u_int32_t next_32(pbuf_t *b)
{
  u_int32_t *a = (u_int32_t *) _next(b, 4);
  return *a;
}

static inline int skip(pbuf_t *b, int n)
{
  b->cur += n;
  assert(b->total >= b->cur);
  return b->cur;
}

static inline int getlen(pbuf_t *b)
{
  return b->total - b->cur;
}

static inline char *getbuf(pbuf_t *b)
{
  return b->p + b->cur;
}

static inline void savepos(pbuf_t *b)
{
  b->saved = b->cur;
}

static inline int getdelta(pbuf_t *b)
{
  return b->cur - b->saved;
}

#define BUF_SIZE 1024
void print_payload(packet_info_t *info)
{
  char   hexbuf  [BUF_SIZE];
  char   charbuf [BUF_SIZE];
  char * buf_ptr;
  char * buf_end;
  int    col;
  int    i;

  printf("-- Payload - len=%d\n", info->plen);
  if(info->plen <= 0 || !info->payload) {
    printf("Nothing to print\n");
    return;
  }

  /* Set the start and end pointers */
  buf_ptr = info->payload;
  buf_end = info->payload + info->plen;

  do {
    col = 0;
    memset(hexbuf, 0, BUF_SIZE);
    memset(charbuf, 0, BUF_SIZE);

    for(i=0;i<18;i++) {
      if(buf_ptr < buf_end) {
	snprintf(hexbuf+(i*3), BUF_SIZE-1,"%.2X ",buf_ptr[0] & 0xFF);

	if(*buf_ptr > 0x1F && *buf_ptr < 0x7E)
	  snprintf(charbuf+i+col, BUF_SIZE-1,"%c",buf_ptr[0]);
	else
	  snprintf(charbuf+i+col, BUF_SIZE-1, ".");

	buf_ptr++;
      }
    }
    printf("%-54s %s\n",hexbuf,charbuf);
  }
  while(buf_ptr < buf_end);
}

int decode_packet(packet_info_t *info, struct pcap_pkthdr *hdr, u_char *pkt)
{
  pbuf_t pbuf;

  assert(info);
  assert(hdr);

  /* reset everything */
  memset(info, 0, sizeof(packet_info_t));

  if(!pkt) return 0;

  pbuf.p = pkt;
  pbuf.cur = 0;
  pbuf.total = hdr->caplen;

  /* get ethernet */
  if(!ether_info(info, &pbuf)) return 0;

  /* IP only */
  if(info->ether.proto != 0x0800) return 0;

  /* get ip */
  if(!ip_info(info, &pbuf)) return 0;

  switch(info->ip.proto) {
  case 6:			/* IPPROTO_TCP */
    if(!tcp_info(info, &pbuf)) return 0;
    break;
  case 17:			/* IPPROTO_UDP */
    if(!udp_info(info, &pbuf)) return 0;
    break;
  case 1:			/* IPPROTO_ICMP */
    if(!icmp_info(info, &pbuf)) return 0;
    break;
  }

  return 1;
}
