/*
   mp_maps.c

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems
   Copyright(C) 2005 John P. Silva, Farm9.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: mp_maps.c,v 1.5 2005/09/02 00:01:25 jsilva Exp $ */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <netdb.h>

#include "mp_conf.h"
#include "mp_maps.h"
#include "mp_util.h"

#define BUFFER_SIZE  8192

static sid_t     *SID_ARRAY;
static int       NSIDS;
static class_t   *CLASS_MAP;
static urlhead_t *URL_MAP;

#define PROTO_NUM 256
static char *proto_names[PROTO_NUM];

/* static prototypes
 */
static void init_proto_names(void);
static void clear_proto_names(void);

typedef void (parsefunc)(char *line, int linenum, char *filename);
static void load_one_map(char *lineprefix, char *file, parsefunc *func);

static void clear_sidmap(void);
static void clear_classmap(void);
static void clear_urlmap(void);

static parsefunc gen_parser;
static parsefunc class_parser;
static parsefunc url_parser;

static int sid_comp(const void *a, const void *b);

static inline void my_free(void *a) {if(a) free(a);}

/* global functions
 */
void init_global_lists()
{
  init_proto_names();
  replace_global_lists(GCF.sid_file, GCF.gen_file, GCF.class_file, GCF.urlhead_file);
}

void replace_global_lists(char *sid_file, char *gen_file, char *class_file, char *urlhead_file)
{
  if(sid_file || gen_file) {
    clear_sidmap();
    load_one_map("1 || ", ((sid_file) ? sid_file : GCF.sid_file), gen_parser);
    load_one_map(NULL, ((gen_file) ? gen_file : GCF.gen_file), gen_parser);
  }
  if(class_file) {
    clear_classmap();
    load_one_map(NULL, class_file, class_parser);
  }
  if(urlhead_file) {
    clear_urlmap();
    load_one_map(NULL, urlhead_file, url_parser);
  }

  if(NSIDS) {
    debug(V_func, "sorting %d sids\n", NSIDS);
    qsort(SID_ARRAY, NSIDS, sizeof(sid_t), sid_comp);
  }
}

void clear_global_lists()
{
  clear_proto_names();
  clear_sidmap();
  clear_classmap();
  clear_urlmap();
}

char *proto_name(int i)
{
  if(i < 0 || i >= PROTO_NUM) return "<unknown>";
  return proto_names[i];
}

/* searches SID array. If nothing found,
   generate fake sid description
 */
sid_t *get_sid(u_long gen, u_long sid)
{
  sid_t search_by = {gen: gen, sid: sid}; /* GCC extension */
  sid_t *tmp;

 again:
  tmp = bsearch(&search_by, SID_ARRAY, NSIDS, sizeof(sid_t), sid_comp);

  /* generate fake sid description and add it to array */
  if(NULL == tmp) {
    char msg[BUFFER_SIZE];
    snprintf(msg, sizeof(msg)-1, "Snort Alert [%u:%u]", gen, sid);
    logprint("fake sid: \"%s\"\n", msg);
    SID_ARRAY = (sid_t *)array_alloc(SID_ARRAY, NSIDS, sizeof(sid_t), 3000);
    SID_ARRAY[NSIDS].gen = gen;
    SID_ARRAY[NSIDS].sid = sid;
    SID_ARRAY[NSIDS].msg = strdup(msg);
    SID_ARRAY[NSIDS].ref = NULL;
    NSIDS++;
    qsort(SID_ARRAY, NSIDS, sizeof(sid_t), sid_comp);

    goto again;			/* find it after qsort */
  }

  return tmp;
}

class_t *get_class(int id)
{
  class_t *ct = CLASS_MAP;

  for(; ct; ct = ct->next)
    if(ct->id == id)
      return ct;

  return NULL;
}

char *get_url(char *buf, int bufsize, char *system, char *id)
{
  char *urlhead = NULL;
  urlhead_t *a;

  assert(buf && system && id);

  for(a = URL_MAP; a; a = a->next)
    if(!strcasecmp(a->system, system)) {
      urlhead = a->url;
      break;
    }

  if(!urlhead) {
    logprint("Reference system '%s' is unknown\n", system);
    return NULL;
  }

  if(snprintf(buf, bufsize - 1, "%s%s", urlhead, id) >= bufsize) {
    logprint("Reference buffer overflow\n");
    return NULL;
  }

  return buf;
}

/* static implementations
 */

static void init_proto_names()
{
  int i;

  for(i = 0; i < PROTO_NUM; i++) {
    struct protoent *pt = getprotobynumber(i);

    if(pt) {
      unsigned char *tmp;
      proto_names[i] = strdup(pt->p_name);

      for(tmp = proto_names[i]; *tmp; tmp++)
	*tmp = (unsigned char) toupper(*tmp);
    }
    else {
      u_char protoname[11];
      snprintf(protoname, 10, "PROTO%03d", i);
      proto_names[i] = strdup(protoname);
    }
  }
}

static void clear_proto_names()
{
  int i;

  for(i = 0; i < PROTO_NUM; i++)
    free(proto_names[i]);
}

/* load_one_map opens the file, reads it line by line, removing newlines,
   spaces, etc (see remove_space() implementation) and calling line-parsing
   function for each line that looks worth it.
   It also remembers to close the file at the end.

   lineprefix parameter is a feature/hack that allows to compensate
   for format differences between sig-msg.map and gen-msg.map files.
*/
static void load_one_map(char *lineprefix, char *file, parsefunc *func)
{
  FILE *fp;
  char buf[BUFFER_SIZE];
  int linenum = 0;
  int buf_size = sizeof(buf) - 1;
  char *buf_pos = buf;
  int prefix_len = 0;

  assert(file);
  assert(func);
  debug(V_func, "-> load_one_map('%s')\n", file);

  if((fp = fopen(file, "r")) == NULL) {
    logerr("open", file);
    return;
  }

  if(lineprefix) {
    prefix_len = strlen(lineprefix);

    /* borrow the space at the beginning of buffer */
    buf_size -= prefix_len; buf_pos += prefix_len;

    /* this should never happen under normal circumstances, I guess */
    if(buf_size < 1)
      fatal("%s:%d: Parser buffer overflow\n", file, linenum + 1);

    /* we want to leave at least 120 chars for incoming data */
    else if(buf_size < 120)
      logprint("%s:%d: the risk of parser buffer overflow\n", file, linenum + 1);
  }

  while(fgets(buf_pos, buf_size, fp)) {
    char *line = remove_space(buf_pos);
    linenum ++;

    if(!*line || *line == '#') {
      debug(V_all, "skipping %s:%d\n", file, linenum);
      continue;
    }

    if(lineprefix)		/* Using the borrowed space... */
      line = memcpy(line - prefix_len, lineprefix, prefix_len);

    func(line, linenum, file);
  }

  debug(V_func, "-> load_one_map(): %d lines processed\n", linenum);

  fclose(fp);
}

static void clear_sidmap()
{
  int i;
  for(i = 0; i < NSIDS; i++) {
    while(SID_ARRAY[i].ref) {
      ref_t *next = SID_ARRAY[i].ref->next;
        
      my_free(SID_ARRAY[i].ref->system);
      my_free(SID_ARRAY[i].ref->id);
      my_free(SID_ARRAY[i].ref);

      SID_ARRAY[i].ref = next;
    }

    my_free(SID_ARRAY[i].msg);
  }

  debug(V_func, "-> clear_sidmap(): cleared %d entries.\n", NSIDS);
  my_free(SID_ARRAY); SID_ARRAY = NULL;
  NSIDS = 0;
}

static void clear_classmap()
{
  static int total = 0;

  if(!CLASS_MAP) {
    debug(V_func, "-> clear_classmap(): cleared %d entries.\n", total);
    total = 0;
  }
  else {
    class_t *next_class = CLASS_MAP->next;

    total ++;
    my_free(CLASS_MAP->type);
    my_free(CLASS_MAP->name);
    my_free(CLASS_MAP);

    CLASS_MAP = next_class;

    clear_classmap();		/* replacing loop by recursion */
  }
}

static void clear_urlmap()
{
  static int total = 0;

  if(!URL_MAP) {
    debug(V_func, "-> clear_urlmap(): cleared %d entries.\n", total);
    total = 0;
  }
  else {
    urlhead_t *next_class = URL_MAP->next;

    total ++;
    if(URL_MAP->system) free(URL_MAP->system);
    if(URL_MAP->url) free(URL_MAP->url);
    free(URL_MAP);

    URL_MAP = next_class;

    clear_urlmap();	     /* calling ourselves again */
  }
}

/* Parses a line from the sid-msg/gen-msg files and
   creates a new entry in SigType chain
 */
static void gen_parser(char *line, int linenum, char *filename)
{
  static char *SEP = "|";
  static char *web_separator = ",";

  int gen = 0, sid = 0;
  char *msg = NULL;

  char *tok;
  int num_toks = 0;
  ref_t *rt = NULL;
  ref_t **rt_addr = &rt;

  assert(line);

  for(tok = strtok(line, SEP); tok && *(tok = remove_space(tok)); tok = strtok(NULL, SEP)) {

    switch(num_toks ++) {
    case 0: gen = strtoul(tok, NULL, 10); break; /* Generator */
    case 1: sid = strtoul(tok, NULL, 10); break; /* Snort ID */
    case 2: msg = tok; break;	/* Comment */

    default: {
      char *system = remove_space(strsep(&tok, web_separator));
      char *id = remove_space(strsep(&tok, web_separator));

      if(!system || !*system || !id || !*id) continue;

      /* we should do just one loop in this while() for each for() loop */
      while(*rt_addr) rt_addr = &(*rt_addr)->next;
      *rt_addr = (ref_t *) my_alloc(sizeof(ref_t));

      debug(V_verb, "%s[%s] | ",
	      (*rt_addr)->system = strdup(system), (*rt_addr)->id = strdup(id));
    }
    } /* switch() */
  }

  if(num_toks < 2) {
    logprint("%s:%d: invalid line\n", filename, linenum);
  }
  else {
    SID_ARRAY = (sid_t *)array_alloc(SID_ARRAY, NSIDS, sizeof(sid_t), 3000);
    debug(V_verb, "(%d/", SID_ARRAY[NSIDS].gen = gen);
    debug(V_verb, "%d) ", SID_ARRAY[NSIDS].sid = sid);
    debug(V_verb, "%s\n", SID_ARRAY[NSIDS].msg = strdup(msg));
    SID_ARRAY[NSIDS].ref = rt;
    NSIDS++;
  }
}

/* parses the context of the 'classification.config' - type file */
static void class_parser(char *line, int linenum, char *filename)
{
  static int class_id = 0;
  char *left = remove_space(strsep(&line, ":"));
  char *tok;
  int num_toks = 0;

  char *type = NULL, *name = NULL;
  int priority = 0;

  class_t **ct = NULL;

  if(!line || !*line || !left || !*left) return;
  else {
    char *config = remove_space(strsep(&left, " \t"));
    char *classification = remove_space(strsep(&left, " \t"));

    if(!config || !*config || !classification || !*classification) return;

    if(strcasecmp(config, "config") || strcasecmp(classification, "classification")) {
      logprint("%s:%d: unrecognized directive\n", filename, linenum);
      return;
    }
  }

  for(tok = strtok(line, ","); tok && *(tok = remove_space(tok)); tok = strtok(NULL, ",")) {

    switch(num_toks ++) {
    case 0: type = tok; break;
    case 1: name = tok; break;
    case 2: priority = strtoul(tok, NULL, 10); break;
    }
  }

  if(num_toks < 3) {
    logprint("%s:%d: invalid line\n", filename, linenum);
    return;
  }

  for(ct = &CLASS_MAP; *ct; ct = &(*ct)->next) {
    if(!strcasecmp((*ct)->type, type)) {
      logprint("%s:%d: duplicate classification '%s'\n", filename, linenum, type);
      return;
    }
  }

  if(!CLASS_MAP) class_id = 0; /* reset class_id if reloading classes */

  *ct = (class_t *) my_alloc(sizeof(class_t));

  debug(V_verb, "(%d)",   (*ct)->id = ++class_id);
  debug(V_verb, " %s ",   (*ct)->type = strdup(type));
  (*ct)->name = strdup(name);
  debug(V_verb, "(%d)\n", (*ct)->priority = priority);
}

/* parses the context of 'reference.config' - type file */
static void url_parser(char *line, int linenum, char *filename)
{
  char *left = remove_space(strsep(&line, ":"));
  char *tok;
  int num_toks = 0;

  char *system = NULL, *url = NULL;

  urlhead_t **uht = NULL;

  if(!line || !*line || !left || !*left) return;
  else {
    char *config = remove_space(strsep(&left, " \t"));
    char *reference = remove_space(strsep(&left, " \t"));

    if(!config || !*config || !reference || !*reference) return;

    if(strcasecmp(config, "config") || strcasecmp(reference, "reference")) {
      logprint("%s:%d: unrecognized directive\n", filename, linenum);
      return;
    }
  }

  for(tok = strtok(line, " "); tok && *(tok = remove_space(tok)); tok = strtok(NULL, " ")) {

    switch(num_toks ++) {
    case 0: system = tok; break;
    case 1: url = tok; break;
    }
  }

  if(num_toks < 2) {
    logprint("%s:%d: invalid line\n", filename, linenum);
    return;
  }

  for(uht = &URL_MAP; *uht; uht = &(*uht)->next) {
    if(!strcasecmp((*uht)->system, system)) {
      logprint("%s:%d: duplicate reference '%s'\n", filename, linenum, system);
      return;
    }
  }

  *uht = (urlhead_t *) my_alloc(sizeof(urlhead_t));

  debug(V_verb, "%s",   (*uht)->system = strdup(system));
  debug(V_verb, "[%s]\n", (*uht)->url = strdup(url));
}

/* for use in qsort/bsearch functions
 */
static int sid_comp(const void *a, const void *b)
{
  return
    100000 * ((int)((sid_t *)a)->gen - (int)((sid_t *)b)->gen) +
    (int)((sid_t *)a)->sid - (int)((sid_t *)b)->sid;
}
