/*
   mp_parser.y

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems
   Copyright(C) 2005 John P. Silva, Farm9.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: mp_parser.y,v 1.3 2005/09/01 02:12:32 jsilva Exp $ */

%{
#define YYSTYPE char *

#define YYDEBUG 3
#define YYERROR_VERBOSE

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <pwd.h>
#include <sys/types.h>
#include <errno.h>

#include "mp_conf.h"
#include "mp_spoolman.h"
#include "mp_util.h"

extern int yylineno;
extern int yylex(void);

static void yyerror(char *msg)
{
/*   printf("line %d: %s\n", yylineno, msg); */
}

#define ASSIGN(x,y) if(x) free(x); x = strdup(y)
static void add_spool(char *dirname);
static void add_module(char *name, char *params);
static void get_uid(char *name);
static char *an(char *);

/*
  variables to save the spool configuration data.
  They are used in add_spool()
  The reason to do it this way is that add_spool is called after
  bison reads closing '}' in the 'spool' statement.
 */
static char *_yy_check_file;
static char *_yy_arch_dir;
static char *_yy_lock;

static char *_yy_class_file;
static char *_yy_sid_file;
static char *_yy_gen_file;
static char *_yy_urlhead_file;

static int   _yy_delete_processed;
static uid_t _yy_uid;
static gid_t _yy_gid;

static out_mod_t *_yy_o_mods;
static out_mod_t **_yy_mod_cur = &_yy_o_mods;
%}

%token ERROR

%token STRING
%token NUM
%token ID

%token T_GLOBAL
%token T_DAEMON
%token T_CF_FILE
%token T_CLASS_FILE
%token T_SID_FILE
%token T_GEN_FILE
%token T_URL_FILE
%token T_PID_FILE
%token T_VERBOSE
%token T_NICE
%token T_RUN_ONCE

%token T_SPOOL
%token T_ARCH_DIR
%token T_CHECK
%token T_DEL_PROC
%token T_OMOD
%token T_USER
%token T_LOCK

%%

 input   : /* empty */
         | input line
         ;

 line    : T_GLOBAL '{' global '}'
         | T_SPOOL STRING { add_spool($2); }
         | T_SPOOL STRING '{'spool'}' { add_spool($2); }

         ;

 global  :  global _global
         | _global
         ;

 spool    : /* empty */
         | spool _spool
         ;

 _global : T_DAEMON                { GCF.daemon_flag = 1; }
         | T_VERBOSE '='    NUM    { GCF.verbose += atoi($3); }
         | T_CLASS_FILE '=' STRING { ASSIGN(GCF.class_file,   $3); }
         | T_SID_FILE '='   STRING { ASSIGN(GCF.sid_file,     $3); }
         | T_GEN_FILE '='   STRING { ASSIGN(GCF.gen_file,     $3); }
         | T_URL_FILE '='   STRING { ASSIGN(GCF.urlhead_file, $3); }
         | T_PID_FILE '='   STRING { ASSIGN(GCF.pid_file,     $3); }
         | T_RUN_ONCE              { GCF.run_once = 1; }
         | T_NICE '='       NUM    { GCF.nice = atoi($3); }
         ;

 _spool  : T_CHECK '='    STRING            { ASSIGN(_yy_check_file, $3); }
         | T_ARCH_DIR '=' STRING            { ASSIGN(_yy_arch_dir, $3); }
         | T_OMOD     '=' STRING            { add_module($3, NULL); }
         | T_OMOD     '=' STRING ',' STRING { add_module($3, $5); }
         | T_USER     '=' STRING            { get_uid($3); }
         | T_LOCK     '=' STRING            { ASSIGN(_yy_lock, an($3)); }
         | T_DEL_PROC                       { _yy_delete_processed = 1; }
         | T_CLASS_FILE '=' STRING          { ASSIGN(_yy_class_file,   $3); }
         | T_SID_FILE '='   STRING          { ASSIGN(_yy_sid_file,     $3); }
         | T_GEN_FILE '='   STRING          { ASSIGN(_yy_gen_file,     $3); }
         | T_URL_FILE '='   STRING          { ASSIGN(_yy_urlhead_file, $3); }
         ;
%%

static void add_spool(char *dirname)
{
  int i = SPOOLS_NUMBER;

  assert(dirname);

  SPOOLS = array_alloc(SPOOLS, SPOOLS_NUMBER++, sizeof(spool_t), 2);

  memset(&SPOOLS[i], 0, sizeof(spool_t));

  SPOOLS[i].spool_dir = strdup(dirname);

  /* those two have been already strdup'ed by ASSIGN() */
  SPOOLS[i].check_file = _yy_check_file;
  SPOOLS[i].arch_dir =   _yy_arch_dir;
  SPOOLS[i].lock =       _yy_lock;

  SPOOLS[i].o_mods = _yy_o_mods;

  SPOOLS[i].delete_processed = _yy_delete_processed;
  SPOOLS[i].uid = _yy_uid;
  SPOOLS[i].gid = _yy_gid;

  SPOOLS[i].class_file =   _yy_class_file;
  SPOOLS[i].sid_file =     _yy_sid_file;
  SPOOLS[i].gen_file =     _yy_gen_file;
  SPOOLS[i].urlhead_file = _yy_urlhead_file;

  /* now reset variables to process the new "spool" statement */
  _yy_class_file = NULL;
  _yy_sid_file = NULL;
  _yy_gen_file = NULL;
  _yy_urlhead_file = NULL;
  _yy_check_file = NULL;
  _yy_arch_dir = NULL;
  _yy_o_mods = NULL;
  _yy_lock = NULL;
  _yy_delete_processed = 0;
  _yy_uid = 0;
  _yy_gid = 0;

  _yy_mod_cur = &_yy_o_mods;
}

static void add_module(char *name, char *params)
{
  assert(name);
  *_yy_mod_cur = my_alloc(sizeof(out_mod_t));

  debug(V_func, "adding module '%s'\n", (*_yy_mod_cur)->name = strdup(name));
  if(params)
    debug(V_func, "adding parameters '%s'\n", (*_yy_mod_cur)->params = strdup(params));

  _yy_mod_cur = &(*_yy_mod_cur)->next;
}

static void get_uid(char *name)
{
  assert(name);

  if(getuid()) {
    logprint("%s:%d: \"user %s\" statement won't work.\n",
	GCF.config_file, yylineno, name);
  }

  else {
    struct passwd *pw = getpwnam(name);

    if(!pw) fatal("%s:%d: '%s': no such user\n",
		  GCF.config_file, yylineno, name);

    _yy_uid = pw->pw_uid;
    _yy_gid = pw->pw_gid;	/* primary group */
  }
}

static char *an(char *name)
{
  char *c = name;
  if(!name) return NULL;
  for( ; *c ; c++)
    if(isalnum(*c) || *c == '_');
    else fatal("%s:%d: \"%s\" contains an illegal symbol '%c'\n",
	       GCF.config_file, yylineno, name, *c);

  return name;
}
