/*
   mp_util.c

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: mp_util.c,v 1.5 2005/09/30 00:21:49 jleplast Exp $ */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <sys/param.h>
#include <paths.h>

#include "mp_conf.h"
#include "mp_util.h"

#define FT "\"%s\": %s\n"

static int I_AM_DAEMON;

/* getting/saving the pid of the parent */
static pid_t _SAVED_PID;

static int make_pidfile(void);

int spool_lock(char *name)
{
  if(!name) return -1;
  else {
    int fd;
    char fil[MAXPATHLEN];

    if(snprintf(fil, sizeof(fil), "%s.%s.lock(%s)",
		_PATH_VARTMP, PACKAGE_NAME, name) > sizeof(fil))
      child_fatal("lock buffer overflow");

    if((fd = open(fil, O_CREAT | O_RDONLY, 0444)) == -1)
      child_fatal("open " FT, fil, strerror(errno));

    if(flock(fd, LOCK_EX) == -1)
      child_fatal("flock " FT, fil, strerror(errno));

    return fd;
  }
}

void spool_unlock(int fd)
{
  if(fd > -1) close(fd);
}

void reset_sigmask(void)
{
  sigset_t set;
  sigemptyset(&set);	/* reset the signal mask */
  sigprocmask(SIG_SETMASK, &set, NULL);
}

/* mudpit is supposedly runs as root and adds dynamic libraries to do stuff.
   We'd better check our uids/gids for suspicious combinations
*/
void check_uids(void)
{
  if(getuid() != geteuid())
    fatal("there should be no reason to SUID this program.\n");
}

pid_t saved_pid(void)
{
  return _SAVED_PID;
}

void save_pid(void)
{
  _SAVED_PID = getpid();
}

/* Check if the file exists and we have an access to it.
   if not, or is its not a file, fatal.
*/
void check_file(char *name, int full_access)
{
  struct stat st;

  assert(name);	       /* there should be no emty dir in this place */
  
  if(name[0] != '/')
    fatal("\"%s\": full path is required.\n", name);

  if(stat(name, &st) == -1)
    fatal(FT, name, strerror(errno));

  if(!S_ISREG(st.st_mode))
    fatal("\"%s\": not a file\n", name);

  if(access(name, full_access ? W_OK : R_OK) == -1)
    fatal(FT, name, strerror(errno));
}

/*
  Check if the directory exists and we have an access to it.
  if not, or is its not a directory, fatal.
 */
void check_dir(char *dir, int full_access)
{
  struct stat st;

  assert(dir);	       /* there should be no emty dir in this place */
  
  if(dir[0] != '/')
    fatal("\"%s\": full path is required.\n", dir);

  if(stat(dir, &st) == -1)
    fatal(FT, dir, strerror(errno));

  if(!S_ISDIR(st.st_mode))
    fatal("\"%s\": not a directory\n", dir);

  if(access(dir, full_access ? R_OK | W_OK | X_OK : R_OK) == -1)
    fatal(FT, dir, strerror(errno));

}

void tv_copy(struct timeval *dst, struct timeval *src)
{
  assert(dst && src);

  dst->tv_sec = src->tv_sec;
  dst->tv_usec = src->tv_usec;
}

void tv_zero(struct timeval *one)
{
  assert (one);
  one->tv_sec = one->tv_usec = 0;
}

/*
  returns:
  0: struct is NOT null
  1: struct is null
*/
int tv_is_null(struct timeval *one)
{
  assert(one);
  return !(one->tv_sec && one->tv_usec);
}

/*
  returns:
  <0: one < two
   0: one == two
  >0: one > two
*/
int tv_compare(struct timeval *one, struct timeval *two)
{
  assert(one && two);

  if(one->tv_sec != two->tv_sec)
    return (one->tv_sec - two->tv_sec);

  return (one->tv_usec - two->tv_usec);
}

double tv_diff(struct timeval *one, struct timeval *two)
{
  assert(one && two);

  return ((float)one->tv_sec + (float)one->tv_usec/1000000.) -
    ((float)two->tv_sec + (float)two->tv_usec/1000000.);
}

/* the main printing facility with several wrappers.
 */
static void _vprint_mes(int priority, const char *format, va_list ap)
{
  assert(priority<=LOG_DEBUG && priority>=LOG_EMERG);
  if(I_AM_DAEMON) {
    vsyslog(priority, format, ap);
  }
  else {
    if(priority<=LOG_ERR) fprintf(stdout,"\n*ERROR*: ");
    vfprintf(stdout, format, ap);
    fflush(stdout);
  }
}

static void _print_mes(int priority, const char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  _vprint_mes(priority, format, ap);
  va_end(ap);
}

void debug(int level, const char *format,...)
{
  va_list ap;
  if(level > GCF.verbose) return;
  va_start(ap, format);
  _vprint_mes(LOG_DEBUG, format, ap);
  va_end(ap);
}

void logprint(const char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  _vprint_mes(LOG_NOTICE, format, ap);
  va_end(ap);
}

void logerr(char *op, char *f)
{
  assert(op);
  if(f) _print_mes(LOG_ERR, "%s "FT, op, f, strerror(errno));
  else _print_mes(LOG_ERR, "%s: %s\n", op, strerror(errno));
}

/* prints error message and kills the whole process group,
   including parent. not exported to an output module.
 */
void fatal(const char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  _vprint_mes(LOG_ERR, format, ap);
  va_end(ap);
  my_exit();
}

/* kills current spool. Do not harm parent.
 */
void child_fatal(const char *format, ...) /* do not kill parent process */
{
  va_list ap;
  va_start(ap, format);
  _vprint_mes(LOG_ERR, format, ap);
  va_end(ap);
  if(saved_pid() != getpid()) exit(1);
}


/* returns the address to the new array's element of size 'elsize'
   allocating/reallocating memory in chunks of 'chunk_size' size
 */
void *array_alloc(void *pointer, int index, int elsize, int chunk_size)
{
#define __ARRAY_CACHE 1
  if(chunk_size <= 0) chunk_size = __ARRAY_CACHE;

  if(!pointer || !(index % chunk_size)) {
    int size = chunk_size * elsize + index * elsize;

    assert(pointer = (pointer) ? realloc(pointer, size) : calloc(1, size));
  }

  return pointer;

#undef __ARRAY_CACHE
}

void *my_alloc(u_long size)
{
  void *p;

  if(!(p = calloc(size, sizeof(char))))
    fatal("Could not allocate %d bytes.\n", size);

  return p;
}

char *remove_space(char *data)
{
  if(data) {
    char *end = data + strlen(data);
    char *start = data;

    while((isspace(*(end - 1))) && end > start ) *(--end) = 0;
    while(isspace(*start)) start++;

    return start;
  }

  else return NULL;
}

void go_daemon(void)	
{
  switch(fork()){
  case -1:			/* error */
    fatal("fork: %s\n", strerror(errno));

  case 0:			/* child */
    save_pid();
    if(!make_pidfile()) my_exit();
    if(!GCF.verbose) {
      int fd = open("/dev/null", O_RDWR);
      if( fd == -1)
	fatal("open " FT, "/dev/null", strerror(errno));
      chdir("/");
      dup2(fd, 0);
      dup2(fd, 1);
      dup2(fd, 2);
      close(fd);
      setsid();
      I_AM_DAEMON = 1;
    }
    break;

  default:			/* parent */
    exit(0);
  }
}

/* returns 1 on success.
   0 otherwise
 */
static int PIDFILE_FD;
static int make_pidfile()
{
  int fd;
  char buffer[20];

  assert(GCF.pid_file);

  if((fd = open(GCF.pid_file, O_CREAT | O_RDWR, 0644)) == -1) {
    logerr("open", GCF.pid_file);
    return 0;
  }

  if(flock(fd, LOCK_EX | LOCK_NB) == -1) {
    logerr("flock", GCF.pid_file);
    close(fd);
    return 0;
  }

  snprintf(buffer, sizeof(buffer) - 1, "%d\n", (int) getpid());
  write(fd, buffer, strlen(buffer));

  /* do not close fd here: this way it will be locked until this process is alive */

  PIDFILE_FD = fd;
  return 1;
}

/* to call from child after fork()
 */
int close_pidfile()
{
  if(PIDFILE_FD)
    close(PIDFILE_FD);

  PIDFILE_FD = 0;
}

void my_exit(void)
{
  if(saved_pid() == getpid())
    kill(0, SIGTERM);

  else exit(CHILD_FATAL);
}

void my_sleep(void)
{
  static int sleep_time = 1;
  debug(V_func, "Sleeping for %d second(s)\n", sleep_time);
  sleep(sleep_time);
}
